# YML 2.2.0 language definition

# written by VB.

import re
from pyPEG import keyword, _and, _not

# pyPEG:
#
#   basestring:     terminal symbol (characters)
#   keyword:        terminal symbol (keyword)
#   matchobj:       terminal symbols (regex, use for scanning symbols)
#   function:       named non-terminal symbol, recursive definition
#                   if you don't want naming in output, precede name with an underscore
#   tuple:          production sequence
#   integer:        count in production sequence:
#                    0: following element is optional
#                   -1: following element can be omitted or repeated endless
#                   -2: following element is required and can be repeated endless
#   list:           options, choose one of them
#   _not:           next element in production sequence is matched only if this would not
#   _and:           next element in production sequence is matched only if this would, too

newSyntax = True

def oldSyntax():
    global newSyntax
    newSyntax = False

def _if(cond, val):
    if cond:
        return val
    else:
        return ()

def listing(x):     return x, -1, (",", x)
r = re.compile

comment = [r(r"//.*"), r(r"/\*.*?\*/", re.S)]
symbol = r(r"[a-zA-Z_](\w|:)*")
pointer = r(r"\*[a-zA-Z_]\w*")
ppointer = r(r"\*\*?[a-zA-Z_]\w*")
macro = r(r"\%[a-zA-Z_]\w*")
reference = r(r"\&[a-zA-Z_]\w*")
xmlSymbol = r(r"[a-zA-Z][\-:a-zA-Z0-9]*")
aliasSymbol = r(r"\-|[a-zA-Z][\-:a-zA-Z0-9]*")
filename = r(r"[^\s;]+")
literal = r(r"""-?\d+\.\d*|-?\.\d+|-?\d+|".*?"|'.*?'""")
ws = r(r"\s+")
def pyExp():        return "!", r(r"(!=|\\!|[^!])+"), "!"
value = [literal, pyExp]

def tagQuote():     return r(r"\].*|\<.*?\>")
def lineQuote():    return r(r"\|.*")
def quote():        return [r(r"\d*>.*"), (literal, 0, [";", "."])]
def parm():         return [([xmlSymbol, pyExp, pointer, macro], "=", [value, pointer]), value, pointer]
_func = [symbol, ppointer, reference], _if(newSyntax, (-1, ("[", listing(parm), "]"))), [("(", listing(parm), ")"), (0, listing(parm))]
def pythonCall():   return keyword("python"), _func, [";", "."]
def declParm():     return [xmlSymbol, pointer, macro], 0, ("=", literal)
def alias():        return keyword("alias"), aliasSymbol
def descend():      return r(r"\+[a-zA-Z][\-:a-zA-Z]*")
def base():         return keyword("is"), symbol
def decl():         return symbol, 0, base, -1, descend, _if(newSyntax, (-1, ("[", 0, listing(declParm), "]"))), 0, ("(", 0, listing(declParm), ")"), 0, alias, 0, content
def python():       return [r(r"!!.*?!!", re.S), r(r"!.*")]
def operator():     return keyword("define"), keyword("operator"), literal, keyword("as"), r(r".*")
def in_ns():        return keyword("in"), xmlSymbol, [_decl, ("{", -2, _decl, "}")]
_decl = keyword("decl"), listing(decl), [";", "."]
def textsection():  return r(r'(\|\|(\>*).*?\|\|(\>*))|(""".*?""")|(\>\>.*?\>\>)', re.S)
def include():      return keyword("include"), 0, reverse, 0, [ktext, kxml], filename, 0, [";", "."]
def func():         return _func, 0, content
def funclist():     return listing(func)
_cmd = funclist, 0, [";", "."]
_inner = [include, textsection, pythonCall, _cmd, quote, lineQuote, tagQuote, pyExp]
def func_plain():   return _func
func_plain.__name__ = "func"
def flist_plain():  return -2, func_plain
flist_plain.__name__ = "funclist"
def xbase():        return flist_plain
def generic():      return flist_plain
def subscript():    return flist_plain
def parentheses():  return "(", 0, funclist, ")"
def fparm():        return flist_plain

_l = _if(newSyntax, ("<", listing(generic), ">"))
_p = _if(not newSyntax, parentheses), _if(newSyntax, ("(", 0, listing(fparm), ")"))
_b = (":", listing(xbase))
_c = [_inner, ("{", -1, _inner, "}")]

def content():      return [ (_l, 0, _p, 0, _b, 0, _c), (_p, 0, _b, 0, _c), (_b, 0, _c), _c ]
def reverse():      return keyword("reverse")
def ktext():        return keyword("text")
def kxml():         return keyword("xml")
def ymlCStyle():    return -1, [_decl, in_ns, include, python, operator, tagQuote, _cmd]
